#include "AuxCompFunctions.h"
#include "KMTFInt.h"

#pragma warning(disable : 4996)

KMTFInt::KMTFInt() {
  lowerBound = upperBound = 0;
  ABC.clear();
  O.clear();
}

void KMTFInt::Init(int k, int l, int h, int first) {
  lowerBound = l;
  upperBound = h;
  kMTF = k;
  FirstEl = first;
  ABC.clear();
  O.clear();
  SetABC();
}

void KMTFInt::SetABC() {
  ABC.clear();
  int lBFlag = 1;
  int uBFlag = 1;

  ABC.push_back(FirstEl);
  int u, l;
  u = FirstEl + 1;
  l = FirstEl - 1;

  while ((lBFlag) || (uBFlag)) {
    if (uBFlag) {
      if (u <= upperBound) {
        ABC.push_back(u++);
      } else {
        uBFlag = 0;
      }
    }

    if (lBFlag) {
      if (l >= lowerBound) {
        ABC.push_back(l--);
      } else {
        lBFlag = 0;
      }
    }
  }
}

int KMTFInt::GetPosition(int v) {
  int i = 0;

  while (v != ABC[i]) {
    i++;
  }

  return i;
}

void KMTFInt::MoveAToFront(vector<int> A) {
  int i, j;
  vector<int> ABCAux;

  ABCAux.clear();
  ABCAux = A;

  int fl;

  for (i = 0; i < ABC.size(); i++) {
    fl = 1;
    j = 0;

    while (fl && (j < (int)A.size())) {
      if (ABC[i] == A[j]) {
        fl = 0;
      } else {
        j++;
      }
    }

    if (fl) {
      ABCAux.push_back(ABC[i]);
    }
  }

  ABC.clear();
  ABC = ABCAux;
}

void KMTFInt::MoveAToFrontOpt(vector<int>& A, std::vector<bool>& lut) {
  int i;
  vector<int> ABCAux;

  ABCAux.clear();
  ABCAux = A;

  for (i = 0; i < ABC.size(); i++) {
    if (!lut[ABC[i]]) {
      ABCAux.push_back(ABC[i]);
    }
  }

  ABC.clear();
  ABC = ABCAux;
}

void KMTFInt::MoveToFront(int p) {
  int a = ABC[p];

  for (int i = p - 1; i >= 0; i--) {
    ABC[i + 1] = ABC[i];
  }

  ABC[0] = a;
}

void KMTFInt::PrintABC(int v) {
  FILE* f = fopen("ABC.TXT", "wt");
  int i;

  fprintf(f, "value %d \n", v);

  for (i = 0; i < ABC.size(); i++) {
    fprintf(f, "i %d, v %d\n", i, ABC[i]);
  }

  fclose(f);
}

vector<int> KMTFInt::DoKMTFInt(vector<int> I) {
  int i, j, position;
  int uv, lv;
  vector<int> A;

  for (i = 1; i < I.size();
       i++) { // the value of the first element is known so we start at 1.
    position = GetPosition(I[i]);
    O.push_back(position);

    if (position != 0) {
      if (position > kMTF) {
        A.clear();
        A.push_back(I[i]);
        uv = I[i] + 1;
        lv = I[i] - 1;

        for (j = 0; j < kMTF; j++) {
          if (uv <= upperBound) {
            A.push_back(uv++);
          }

          if (lv >= 0) {
            A.push_back(lv--);
          }
        }

        MoveAToFront(A);
      } else {
        MoveToFront(position);
      }
    }
  }

  return O;
}

vector<int> KMTFInt::DoKMTFIntOpt(vector<int> I) {
  std::vector<bool> lut;
  lut.reserve(256);

  int i, j, position;
  int uv, lv;
  vector<int> A;

  for (i = 1; i < I.size();
       i++) { // the value of the first element is known so we start at 1.
    position = GetPosition(I[i]);
    O.push_back(position);

    if (position != 0) {
      if (position > kMTF) {
        std::fill(lut.begin(), lut.end(), false);

        A.clear();
        A.push_back(I[i]);

        lut[I[i]] = true;

        uv = I[i] + 1;
        lv = I[i] - 1;

        for (j = 0; j < kMTF; j++) {
          if (uv <= upperBound) {
            A.push_back(uv++);
            lut[uv - 1] = true;
          }

          if (lv >= 0) {
            A.push_back(lv--);
            lut[lv + 1] = true;
          }
        }

        MoveAToFrontOpt(A, lut);
      } else {
        MoveToFront(position);
      }
    }
  }

  return O;
}

vector<int> KMTFInt::DoDeKMTFInt(vector<int> I) {
  int i, j, position, value;
  int uv, lv;
  vector<int> A;

  O.push_back(FirstEl);

  for (i = 0; i < I.size(); i++) {
    value = ABC[I[i]];
    O.push_back(value);
    position = I[i];

    if (position != 0) {
      if (position > kMTF) {
        A.clear();
        A.push_back(value);
        uv = value + 1;
        lv = value - 1;

        for (j = 0; j < kMTF; j++) {
          if (uv <= upperBound) {
            A.push_back(uv++);
          }

          if (lv >= 0) {
            A.push_back(lv--);
          }
        }

        MoveAToFront(A);
      } else {
        MoveToFront(position);
      }
    }
  }

  return O;
}

vector<int> KMTFInt::DoDeKMTFIntOpt(vector<int> I) {
  std::vector<bool> lut;
  lut.reserve(256);

  int i, j, position, value;
  int uv, lv;
  vector<int> A;

  O.push_back(FirstEl);

  for (i = 0; i < I.size(); i++) {
    value = ABC[I[i]];
    O.push_back(value);
    position = I[i];

    if (position != 0) {
      if (position > kMTF) {
        std::fill(lut.begin(), lut.end(), false);

        A.clear();
        A.push_back(value);

        lut[value] = true;

        uv = value + 1;
        lv = value - 1;

        for (j = 0; j < kMTF; j++) {
          if (uv <= upperBound) {
            A.push_back(uv++);
            lut[uv - 1] = true;
          }

          if (lv >= 0) {
            A.push_back(lv--);
            lut[lv + 1] = true;
          }
        }

        MoveAToFrontOpt(A, lut);
      } else {
        MoveToFront(position);
      }
    }
  }

  return O;
}
