#include <stdio.h>
#include <string.h>

#include <vector>

#include "AuxCompFunctions.h"
#include "KMTFInt.h"
#include "MatToVec.h"
#include "MwI.h"
#include "MyCPUTime.h"
#include "MyRawImageData.h"

#pragma warning(disable : 4996)

using namespace std;

double HBatchSum = 0;

void DetermineRange(int& lowerBound, int& upperBound, vector<int> I) {
  lowerBound = 100000;
  upperBound = 0;

  for (int i = 0; i < I.size(); i++) {
    if (I[i] > upperBound) {
      upperBound = I[i];
    } else if (I[i] < lowerBound) {
      lowerBound = I[i];
    }
  }
}

void TransforWith_k_MTF(vector<int>& V, int K_range) {
  MyCPUTime t1;

  t1.Start();

  KMTFInt KMTF;
  int first = V[0];
  int l, h;

  DetermineRange(l, h, V);

  KMTF.Init(K_range, l, h, first);

  vector<int> VKMTF = KMTF.DoKMTFInt(V);

  t1.Stop();
  double d = t1.GetEllapsedCPUTime();
  printf("CPU Time %f\n", d);

  double HkMTF = CalculateEntropy(DoVectorStatistic(VKMTF));
  printf("\n ------------- k-MTF ---------------\n");
  printf(" H MTF: %f \n", HkMTF);

  KMTFInt DeKMTF;
  DeKMTF.Init(K_range, l, h, first);
  vector<int> DeKMTFO = DeKMTF.DoDeKMTFInt(VKMTF);

  if (V == DeKMTFO) {
    printf(" DeMTF OK.");
  } else {
    printf(" ------> DeMTF failed.");
  }
}

void TransforWith_k_MTFOpt(vector<int>& V, int K_range) {
  MyCPUTime t1;

  t1.Start();

  KMTFInt KMTF;
  int first = V[0];
  int l, h;

  DetermineRange(l, h, V);

  KMTF.Init(K_range, l, h, first);

  vector<int> VKMTF = KMTF.DoKMTFIntOpt(V);

  t1.Stop();
  double d = t1.GetEllapsedCPUTime();
  printf("CPU Time %f\n", d);

  double HkMTF = CalculateEntropy(DoVectorStatistic(VKMTF));
  printf("\n ------------- k-MTF ---------------\n");
  printf(" H MTF: %f \n", HkMTF);

  KMTFInt DeKMTF;
  DeKMTF.Init(K_range, l, h, first);
  vector<int> DeKMTFO = DeKMTF.DoDeKMTFIntOpt(VKMTF);

  if (V == DeKMTFO) {
    printf(" DeMTF OK.");
  } else {
    printf(" ------> DeMTF failed.");
  }
}

void TransformWithMwI(vector<int>& V, int Delta) {
  MyCPUTime t1;

  t1.Start();

  int first = V[0];
  int l, h;

  DetermineRange(l, h, V);
  MwI mvi;
  mvi.Init(Delta, l, h, first);
  vector<int> VMwI = mvi.DoMwI(V);

  t1.Stop();
  double d = t1.GetEllapsedCPUTime();
  printf("CPU Time %f\n", d);

  double H = CalculateEntropy(DoVectorStatistic(VMwI));
  printf("\n ------------- MwI ---------------\n");
  printf(" H MwI: %f \n", H);

  MwI DeMwI;
  DeMwI.Init(Delta, l, h, first);

  vector<int> DeMwIV = DeMwI.DoDeMwI(VMwI);

  if (V == DeMwIV) {
    printf(" DeMwI OK.");
  } else {
    printf(" ------> DeMwI failed.");
    int i = 0;

    while ((i < DeMwIV.size()) && (DeMwIV[i] == V[i])) {
      i++;
    }
    printf("-----> Error at index %d \n", i);
  }
}

/////////////////// Batch

double TransforWithMwIBatch(MyRawImageData& image, vector<int>& V, int delta) {
  MwI MWI;
  int first = V[0];
  int l, h;

  MyCPUTime t1;
  t1.Start();

  DetermineRange(l, h, V);

  MWI.Init(delta, l, h, first);

  vector<int> VMwI = MWI.DoMwI(V);

  t1.Stop();
  double d = t1.GetEllapsedCPUTime();

  double HMwI = CalculateEntropy(DoVectorStatistic(VMwI));
  printf("H: %f\tCPU Time: %0.3f\t%s\n", HMwI, d, image.GetLastLoaded().c_str());
  HBatchSum += HMwI;

  return d;
}

double TransforWith_k_MTFBatch(MyRawImageData& image, vector<int>& V, int delta) {
  KMTFInt KMTF;
  int first = V[0];
  int l, h;

  MyCPUTime t1;
  t1.Start();

  DetermineRange(l, h, V);

  KMTF.Init(delta, l, h, first);

  vector<int> VKMTF = KMTF.DoKMTFInt(V);

  t1.Stop();
  double d = t1.GetEllapsedCPUTime();

  double HkMTF = CalculateEntropy(DoVectorStatistic(VKMTF));
  printf("H: %f\tCPU Time: %0.3f\t%s\n", HkMTF, d, image.GetLastLoaded().c_str());
  HBatchSum += HkMTF;

  return d;
}

double TransforWith_k_MTFOptBatch(MyRawImageData& image, vector<int>& V, int delta) {
  KMTFInt KMTF;
  int first = V[0];
  int l, h;

  MyCPUTime t1;
  t1.Start();

  DetermineRange(l, h, V);

  KMTF.Init(delta, l, h, first);

  vector<int> VKMTF = KMTF.DoKMTFIntOpt(V);

  t1.Stop();
  double d = t1.GetEllapsedCPUTime();

  double HkMTF = CalculateEntropy(DoVectorStatistic(VKMTF));
  printf("H: %f\tCPU Time: %0.3f\t%s\n", HkMTF, d, image.GetLastLoaded().c_str());
  HBatchSum += HkMTF;

  return d;
}

void ProcessBatch(MyRawImageData& MRID) {
  int sw = 0;

  printf("Transforms:\n");
  printf("1...Implementation with vector\n");
  printf("2...Implementation with linked list\n");
  printf("3...Implementation with lookup table\n");

  int trs;
  scanf("%d", &trs);

  int delta = 12;

  vector<vector<int>> Image;
  vector<int> VI;
  MatToVec MtoV;
  double d = 0.0;

  while (MRID.ReturnNextBatchData(Image) == 1) {
    MtoV.SetMatrix(Image);
    MtoV.ScanLine();
    VI = MtoV.ReturnVector();

    switch (trs) {
      case 1:
        d += TransforWith_k_MTFBatch(MRID, VI, delta);
        break;

      case 2:
        d += TransforWithMwIBatch(MRID, VI, delta);
        break;

      case 3:
        d += TransforWith_k_MTFOptBatch(MRID, VI, delta);
        break;

      default:
        break;
    }
  }
  int NoOfBatches = MRID.ReturnNoOfBatches();
  printf("------\n");
  printf("Hsum: %f\n", HBatchSum);
  double average = HBatchSum / (double)NoOfBatches;
  printf("average H: %0.3f\n", average);
  printf("CPU Time(sum): %f\n", d);
}

int main() {
  MyRawImageData MRID;

  int ImageIndex = MRID.SelectImage();

  if (ImageIndex == 99) {
    return 1;
  }

  if (ImageIndex == 0) {
    ProcessBatch(MRID);
  } else {
    if (MRID.LoadBMP()) {
      vector<vector<int>> Image = MRID.ReturnImageData();
      printf("Original H: %f \n", MRID.ReturnEntropy());

      vector<int> VI;
      MatToVec MtoV;

      MtoV.SetMatrix(Image);
      MtoV.ScanLine();
      VI = MtoV.ReturnVector();
      MtoV.SetVector(VI);

      double StDev = CalculateStandardDeviation(VI);
      printf("Standard deviation %f\n", StDev);

      int DeOvFlag = 0;
      DeOvFlag = MtoV.DeScanLine();

      if (DeOvFlag) {
        printf("->V to M is rigth. \n");
      } else {
        printf("-----> V to M failed. \n");
      }

      printf("Transforms:\n");
      printf("1...Implementation with vector\n");
      printf("2...Implementation with linked list\n");
      printf("3...Implementation with lookup table\n");

      int trs;
      scanf("%d", &trs);

      int delta = 12;

      switch (trs) {
        case 1:
          TransforWith_k_MTF(VI, delta);
          break;

        case 2:
          TransformWithMwI(VI, delta);
          break;

        case 3:
          TransforWith_k_MTFOpt(VI, delta);
          break;

        default:
          break;
      }
    }
  }

  return 0;
}
