package gemma;

import java.util.Arrays;
import java.util.Collections;
import java.util.Vector;

public class KMTFInt implements Cpp {

  private int lowerBound;
  private int upperBound;
  private int kMTF;
  private int FirstEl;
  private Vector<Integer> ABC;
  private Vector<Integer> O;

  public void Init(int k, int l, int h, int first) {
    lowerBound = l;
    upperBound = h;
    kMTF = k;
    FirstEl = first;
    O = new Vector<Integer>();
    SetABC();
  }

  private void SetABC() {
    ABC = new Vector<Integer>();
    int lBFlag = 1;
    int uBFlag = 1;

    ABC.add(FirstEl);
    int u, l;
    u = FirstEl + 1;
    l = FirstEl - 1;

    while ((lBFlag != 0) || (uBFlag != 0)) {
      if (uBFlag != 0) {
        if (u <= upperBound) {
          ABC.add(u++);
        } else {
          uBFlag = 0;
        }
      }

      if (lBFlag != 0) {
        if (l >= lowerBound) {
          ABC.add(l--);
        } else {
          lBFlag = 0;
        }
      }
    }
  }

  private int GetPosition(int v) {
    int i = 0;

    while (v != ABC.get(i).intValue()) i++;

    return i;
  }

  private void MoveAToFront(Vector<Integer> A) {
    int i;
    Vector<Integer> ABCAux = new Vector<Integer>(A);

    for (i = 0; i < ABC.size(); i++) {
      if (!A.contains(ABC.get(i))) {
        ABCAux.add(ABC.get(i));
      }
    }
    ABC = ABCAux;
  }

  private void MoveAToFrontOpt(Vector<Integer> A, Vector<Boolean> lut) {
    int i;
    Vector<Integer> ABCAux = new Vector<Integer>(A);

    for (i = 0; i < ABC.size(); i++) {
      if (!lut.get(ABC.get(i))) {
        ABCAux.add(ABC.get(i));
      }
    }
    ABC = ABCAux;
  }

  private void MoveToFront(int p) {
    int a = ABC.get(p);

    for (int i = p - 1; i >= 0; i--) {
      ABC.set(i + 1, ABC.get(i));
    }

    ABC.set(0, a);
  }

  public Vector<Integer> DoKMTFInt(Vector<Integer> I) {

    int i, j, position;
    int uv, lv;
    Vector<Integer> A = new Vector<Integer>();
    for (i = 1; i < I.size(); i++) // the value of the first element is known so we start at 1.
    {
      //	PrintABC(I[i]);

      int val = I.get(i);

      position = GetPosition(val);
      O.add(position);
      if (position != 0) {

        if (position > kMTF) {
          A.clear();
          A.add(val);
          uv = val + 1;
          lv = val - 1;
          for (j = 0; j < kMTF; j++) {
            if (uv <= upperBound) {
              A.add(uv++);
            }
            if (lv >= 0) {
              A.add(lv--);
            }
          }

          MoveAToFront(A);
        } else {
          MoveToFront(position);
        }
      }
    }

    return O;
  }

  public Vector<Integer> DoKMTFIntOpt(Vector<Integer> I) {

    Vector<Boolean> lut = new Vector<Boolean>();
    lut.setSize(256);
    int i, j, position;
    int uv, lv;
    Vector<Integer> A = new Vector<Integer>();
    for (i = 1; i < I.size(); i++) // the value of the first element is known so we start at 1.
    {
      int val = I.get(i);

      position = GetPosition(val);
      O.add(position);
      if (position != 0) {

        if (position > kMTF) {
          Collections.fill(lut, false);
          A.clear();
          A.add(val);
          lut.set(val, true);
          uv = val + 1;
          lv = val - 1;
          for (j = 0; j < kMTF; j++) {
            if (uv <= upperBound) {
              A.add(uv++);
              lut.set(uv - 1, true);
            }
            if (lv >= 0) {
              A.add(lv--);
              lut.set(lv + 1, true);
            }
          }

          MoveAToFrontOpt(A, lut);
        } else {
          MoveToFront(position);
        }
      }
    }

    return O;
  }

  public Vector<Integer> DoDeKMTFInt(Vector<Integer> I) {

    int i, j, position, value;
    int uv, lv;
    Vector<Integer> A = new Vector<>();

    O.add(FirstEl);

    for (i = 0; i < I.size(); i++) {
      value = ABC.get(I.get(i));
      O.add(value);
      position = I.get(i);

      if (position != 0) {

        if (position > kMTF) {
          A.clear();
          A.add(value);
          uv = value + 1;
          lv = value - 1;
          for (j = 0; j < kMTF; j++) {
            if (uv <= upperBound) A.add(uv++);
            if (lv >= 0) A.add(lv--);
          }

          MoveAToFront(A);
        } else MoveToFront(position);
      }
    }

    return O;
  }

  public Vector<Integer> DoDeKMTFIntOpt(Vector<Integer> I) {

    Vector<Boolean> lut = new Vector<Boolean>();
    lut.setSize(256);
    int i, j, position, value;
    int uv, lv;
    Vector<Integer> A = new Vector<>();

    O.add(FirstEl);

    for (i = 0; i < I.size(); i++) {
      value = ABC.get(I.get(i));
      O.add(value);
      position = I.get(i);

      if (position != 0) {

        if (position > kMTF) {
          Collections.fill(lut, false);
          A.clear();
          A.add(value);
          lut.set(value, true);
          uv = value + 1;
          lv = value - 1;
          for (j = 0; j < kMTF; j++) {
            if (uv <= upperBound) {
              A.add(uv++);
              lut.set(uv - 1, true);
            }
            if (lv >= 0) {
              A.add(lv--);
              lut.set(lv + 1, true);
            }
          }

          MoveAToFrontOpt(A, lut);
        } else {
          MoveToFront(position);
        }
      }
    }

    return O;
  }
}
