package gemma;

import java.util.Vector;

public class Main implements Cpp {

  private double HBatchSum = 0;

  private int[] DetermineRange(Vector<Integer> I) {
    int lowerBound = 100000;
    int upperBound = 0;

    for (int i = 0; i < I.size(); i++) {
      if (I.get(i) > upperBound) upperBound = I.get(i);
      else if (I.get(i) < lowerBound) lowerBound = I.get(i);
    }

    return new int[] {lowerBound, upperBound};
  }

  private double TransforWithMwIBatch(
      final RawImageData image, final Vector<Integer> V, int delta) {
    MwI MWI = new MwI();
    int first = V.get(0);
    int l, h;

    CPUTime t1 = new CPUTime();
    t1.Start();

    int[] lh = DetermineRange(V);
    l = lh[0];
    h = lh[1];

    MWI.Init(delta, l, h, first);

    Vector<Integer> VMwI = MWI.DoMwI(V);

    t1.Stop();
    double d = t1.GetEllapsedCPUTime();

    double HMwI = AuxCompFunctions.CalculateEntropy(AuxCompFunctions.DoVectorStatistic(VMwI));
    printf("H: %f\tCPU Time: %0.3f\t%s\n", HMwI, d, image.GetLastLoaded());
    HBatchSum += HMwI;

    return d;
  }

  private double TransforWith_k_MTFBatch(
      final RawImageData image, final Vector<Integer> V, int delta) {
    KMTFInt KMTF = new KMTFInt();
    int first = V.get(0);
    int l, h;

    CPUTime t1 = new CPUTime();
    t1.Start();

    int[] lh = DetermineRange(V);
    l = lh[0];
    h = lh[1];

    KMTF.Init(delta, l, h, first);

    Vector<Integer> VKMTF = KMTF.DoKMTFInt(V);

    t1.Stop();
    double d = t1.GetEllapsedCPUTime();

    double HKMTF = AuxCompFunctions.CalculateEntropy(AuxCompFunctions.DoVectorStatistic(VKMTF));
    printf("H: %f\tCPU Time: %0.3f\t%s\n", HKMTF, d, image.GetLastLoaded());
    HBatchSum += HKMTF;

    return d;
  }

  private double TransforWith_k_MTF_OptBatch(
      final RawImageData image, final Vector<Integer> V, int delta) {
    KMTFInt KMTF = new KMTFInt();
    int first = V.get(0);
    int l, h;

    CPUTime t1 = new CPUTime();
    t1.Start();

    int[] lh = DetermineRange(V);
    l = lh[0];
    h = lh[1];

    KMTF.Init(delta, l, h, first);

    Vector<Integer> VKMTF = KMTF.DoKMTFIntOpt(V);

    t1.Stop();
    double d = t1.GetEllapsedCPUTime();

    double HKMTF = AuxCompFunctions.CalculateEntropy(AuxCompFunctions.DoVectorStatistic(VKMTF));
    printf("H: %f\tCPU Time: %0.3f\t%s\n", HKMTF, d, image.GetLastLoaded());
    HBatchSum += HKMTF;

    return d;
  }

  private void TransformWithMwI(Vector<Integer> V, int Delta) {

    CPUTime t1 = new CPUTime();
    t1.Start();

    int first = V.get(0);
    int l, h;

    int[] lh = DetermineRange(V);
    l = lh[0];
    h = lh[1];

    MwI mvi = new MwI();
    mvi.Init(Delta, l, h, first);
    Vector<Integer> VMwI = mvi.DoMwI(V);

    t1.Stop();
    double d = t1.GetEllapsedCPUTime();
    printf("CPU Time %f\n", d);

    double H = AuxCompFunctions.CalculateEntropy(AuxCompFunctions.DoVectorStatistic(VMwI));
    printf("\n ------------- MwI ---------------\n");
    printf(" H MwI: %f \n", H);

    MwI DeMwI = new MwI();
    DeMwI.Init(Delta, l, h, first);

    Vector<Integer> DeMwIV = DeMwI.DoDeMwI(VMwI);
    if (V.equals(DeMwIV)) {
      printf(" DeMwI OK.");
    } else {
      printf(" ------> DeMwI failed.");
      int i = 0;
      while ((i < DeMwIV.size()) && (DeMwIV.get(i).equals(V.get(i)))) {
        i++;
      }
      printf("-----> Error at index %d \n", i);
    }
  }

  private void TransforWith_k_MTF(final Vector<Integer> V, final int K_range) {
    CPUTime t1 = new CPUTime();
    t1.Start();

    KMTFInt KMTF = new KMTFInt();
    int first = V.get(0);
    int l, h;

    int[] lh = DetermineRange(V);
    l = lh[0];
    h = lh[1];

    KMTF.Init(K_range, l, h, first);

    Vector<Integer> VKMTF = KMTF.DoKMTFInt(V);

    t1.Stop();
    double d = t1.GetEllapsedCPUTime();
    printf("CPU Time %f\n", d);

    double HkMTF = AuxCompFunctions.CalculateEntropy(AuxCompFunctions.DoVectorStatistic(VKMTF));
    printf("\n ------------- k-MTF ---------------\n");
    printf(" H MTF: %f \n", HkMTF);

    KMTFInt DeKMTF = new KMTFInt();
    DeKMTF.Init(K_range, l, h, first);
    Vector<Integer> DeKMTFO = DeKMTF.DoDeKMTFInt(VKMTF);
    if (V.equals(DeKMTFO)) {
      printf(" DeMTF OK.");
    } else {
      printf(" ------> DeMTF failed.");
    }
  }

  private void TransforWith_k_MTF_Opt(final Vector<Integer> V, final int K_range) {
    CPUTime t1 = new CPUTime();
    t1.Start();

    KMTFInt KMTF = new KMTFInt();
    int first = V.get(0);
    int l, h;

    int[] lh = DetermineRange(V);
    l = lh[0];
    h = lh[1];

    KMTF.Init(K_range, l, h, first);

    Vector<Integer> VKMTF = KMTF.DoKMTFIntOpt(V);

    t1.Stop();
    double d = t1.GetEllapsedCPUTime();
    printf("CPU Time %f\n", d);

    double HkMTF = AuxCompFunctions.CalculateEntropy(AuxCompFunctions.DoVectorStatistic(VKMTF));
    printf("\n ------------- k-MTF ---------------\n");
    printf(" H MTF: %f \n", HkMTF);

    KMTFInt DeKMTF = new KMTFInt();
    DeKMTF.Init(K_range, l, h, first);
    Vector<Integer> DeKMTFO = DeKMTF.DoDeKMTFInt(VKMTF);
    if (V.equals(DeKMTFO)) {
      printf(" DeMTF OK.");
    } else {
      printf(" ------> DeMTF failed.");
    }
  }

  private void ProcessBatch(RawImageData MRID) {

    printf("Transforms:\n");
    printf("1...Implementation with vector\n");
    printf("2...Implementation with linked list\n");
    printf("3...Implementation with lookup table\n");

    int trs = scanf(Integer.class);

    Vector<Vector<Integer>> Image = new Vector<Vector<Integer>>();
    Vector<Integer> VI;
    MatToVec MtoV = new MatToVec();
    double d = 0.0;

    while (MRID.ReturnNextBatchData(Image) == 1) {
      MtoV.SetMatrix(Image);
      MtoV.ScanLine();
      VI = MtoV.ReturnVector();

      switch (trs) {
        case 1:
          d += TransforWith_k_MTFBatch(MRID, VI, 12);
          break;
        case 2:
          d += TransforWithMwIBatch(MRID, VI, 12);
          break;
        case 3:
          d += TransforWith_k_MTF_OptBatch(MRID, VI, 12);
          break;
        default:
          break;
      }
      Image.clear();
    }
    int NoOfBatches = MRID.ReturnNoOfBatches();
    printf("------\n");
    printf("Hsum: %f\n", HBatchSum);
    double average = HBatchSum / (double) NoOfBatches;
    printf("average H: %0.3f\n", average);
    printf("CPU Time(sum): %f\n", d);
  }

  public void execute() {

    RawImageData MRID = new RawImageData();

    int ImageIndex = MRID.SelectImage();

    if (ImageIndex == 99) {
      return;
    }

    if (ImageIndex == 0) {
      ProcessBatch(MRID);
    } else {

      if (MRID.LoadBMP() != 0) {
        Vector<Vector<Integer>> Image = MRID.ReturnImageData();
        printf("Original H: %f \n", MRID.ReturnEntropy());

        Vector<Integer> VI;
        MatToVec MtoV = new MatToVec();

        MtoV.SetMatrix(Image);
        MtoV.ScanLine();

        VI = MtoV.ReturnVector();
        MtoV.SetVector(VI);

        double StDev = AuxCompFunctions.CalculateStandardDeviation(VI);
        printf("Standard deviation %f\n", StDev);

        printf("Transforms:\n");
        printf("1...Implementation with vector\n");
        printf("2...Implementation with linked list\n");
        printf("3...Implementation with lookup table\n");

        int trs = scanf(Integer.class);

        switch (trs) {
          case 1:
            TransforWith_k_MTF(VI, 12);
            break;
          case 2:
            TransformWithMwI(VI, 12);
            break;
          case 3:
            TransforWith_k_MTF_Opt(VI, 12);
            break;
          default:
            break;
        }
      }
    }

    printf("\n\n");
  }

  public static void main(String[] args) {
    new Main().execute();
  }
}
