package gemma;

import java.util.Vector;

public class MwI implements Cpp {

  public static class ABCElement {
    public int el;
    public ABCElement next;
  }

  private int lowerBound;
  private int upperBound;
  private int ABCSize;
  private int delta;
  private int FirstEl;
  private ABCElement ABC;
  private ABCElement ABCTail;
  private ABCElement Pool;
  private Vector<Integer> O;

  public void Init(int k, int l, int h, int first) {
    lowerBound = l;
    upperBound = h;
    ABCSize = h - l + 1;
    delta = k;
    FirstEl = first;

    O = new Vector<Integer>();
    ABCElement abc;
    for (int i = 0; i < 2 * ABCSize; i++) {
      abc = new ABCElement();
      abc.el = 0;
      AddToPool(abc);
    }

    InitABC();
    //		PrintABC();
  }

  private void AddToPool(ABCElement el) {
    el.next = Pool;
    Pool = el;
  }

  private ABCElement GetFromPool() {
    ABCElement p = Pool;
    Pool = Pool.next;
    return p;
  }

  private void InitABC() {

    int lBFlag = 1;
    int uBFlag = 1;
    int ABCIndex = 0;

    ABCElement p = GetFromPool();
    p.el = FirstEl;
    p.next = null;
    ABC = ABCTail = p;

    int u, l;
    u = FirstEl + 1;
    l = FirstEl - 1;
    ABCIndex++;
    while ((lBFlag != 0) || (uBFlag != 0)) {
      if (uBFlag != 0) {
        p = GetFromPool();
        p.el = u++;
        p.next = null;
        ABCTail.next = p;
        ABCTail = p;

        if (u > upperBound) uBFlag = 0;
      }

      if (lBFlag != 0) {
        p = GetFromPool();
        p.el = l--;
        p.next = null;
        ABCTail.next = p;
        ABCTail = p;

        if (l < lowerBound) lBFlag = 0;
      }
    }
  }

  private int GetPosition(int v, final ref<ABCElement> qPrev, final ref<ABCElement> q) {
    if (v == ABC.el) return 0;

    int i = 1;
    qPrev.item = ABC;
    q.item = qPrev.item.next;

    while (q.item.el != v) {
      i++;
      qPrev.item = q.item;
      q.item = q.item.next;
    }

    return i;
  }

  public Vector<Integer> DoMwI(Vector<Integer> I) {

    O.setSize(I.size() - 1);
    int Oindex = 0;

    int i, j, position;
    int uv, lv;

    ABCElement T, Ttail;
    ABCElement p;
    ABCElement qPrev = null, q = null;

    for (i = 1; i < I.size(); i++) // the value of the first element is known so we start at 1.
    {
      //    	printf("%d\n",i);
      ref<ABCElement> _qPrev = new ref<>(qPrev);
      ref<ABCElement> _q = new ref<>(q);

      position = GetPosition(I.get(i), _qPrev, _q);
      O.set(Oindex++, position);

      qPrev = _qPrev.item;
      q = _q.item;

      if (position != 0) {

        if (position > delta) {
          p = GetFromPool();
          p.el = I.get(i);
          p.next = null;
          T = Ttail = p;

          uv = I.get(i) + 1;
          lv = I.get(i) - 1;
          for (j = 0; j < delta; j++) {
            if (uv <= upperBound) {
              p = GetFromPool();
              p.el = uv++;
              p.next = null;
              Ttail.next = p;
              Ttail = p;
            }

            if (lv >= lowerBound) {
              p = GetFromPool();
              p.el = lv--;
              p.next = null;
              Ttail.next = p;
              Ttail = p;
            }
          }

          MoveTToFront(T, Ttail, lv + 1, uv - 1);
        } else // move element to front
        {
          qPrev.next = q.next;
          q.next = ABC;
          ABC = q;
        }
      }
    }

    return O;
  }

  private void MoveTToFront(final ABCElement head, final ABCElement tail, int a, int b) {
    // first remove the elements from interval [a, b] from ABC

    //  	PrintList(head);

    ABCElement p1, p2;
    int elToRemove = b - a + 1;
    int elindex = 0;

    while ((ABC != null)
        && (elindex < elToRemove)
        && (ABC.el >= a)
        && (ABC.el <= b)) { // first element should be removed
      p1 = ABC;
      ABC = ABC.next;
      AddToPool(p1);
      elindex++;
    }

    if (ABC != null) { // elements in between should be removed

      p1 = ABC;
      p2 = p1.next;

      while ((p2 != null) && (elindex < elToRemove)) {
        if ((p2.el < a) || (p2.el > b)) // out of range, the element remains
        {
          p1 = p2;
          p2 = p2.next;
        } else // remove the element
        {
          p1.next = p2.next;
          AddToPool(p2);
          p2 = p1.next;
          elindex++;
        }
      }
    }
    // now append the lest T to ABC

    tail.next = ABC;
    ABC = head;
  }

  private int GetValue(int position, final ref<ABCElement> qPrev, final ref<ABCElement> q) {
    if (position == 0) {
      return ABC.el;
    }

    qPrev.item = ABC;
    q.item = qPrev.item.next;
    int i = 1;

    while (i != position) {
      i++;
      qPrev.item = q.item;
      q.item = q.item.next;
    }
    return q.item.el;
  }

  public Vector<Integer> DoDeMwI(Vector<Integer> I) {
    O.clear();
    O.setSize(I.size() + 1);

    int Oindex = 0;

    int i, j, position;
    int value;
    int uv, lv;

    ABCElement T, Ttail;
    ABCElement p;
    ABCElement qPrev = null, q = null;

    O.set(Oindex++, FirstEl);

    for (i = 0; i < I.size(); i++) {
      ref<ABCElement> _qPrev = new ref<>(qPrev);
      ref<ABCElement> _q = new ref<>(q);

      position = I.get(i);
      value = GetValue(position, _qPrev, _q);
      O.set(Oindex++, value);

      qPrev = _qPrev.item;
      q = _q.item;

      if (position != 0) {

        if (position > delta) {
          p = GetFromPool();
          p.el = value;
          p.next = null;
          T = Ttail = p;

          uv = value + 1;
          lv = value - 1;
          for (j = 0; j < delta; j++) {
            if (uv <= upperBound) {
              p = GetFromPool();
              p.el = uv++;
              p.next = null;
              Ttail.next = p;
              Ttail = p;
            }

            if (lv >= lowerBound) {
              p = GetFromPool();
              p.el = lv--;
              p.next = null;
              Ttail.next = p;
              Ttail = p;
            }
          }

          MoveTToFront(T, Ttail, lv + 1, uv - 1);
        } else // move element to front
        {
          qPrev.next = q.next;
          q.next = ABC;
          ABC = q;
        }
      }
    }
    return O;
  }
}
